/*
 * COPYRIGHT:        See COPYRIGHT.TXT
 * PROJECT:          Ext2 File System Driver for WinNT/2K/XP
 * FILE:             create.c
 * PROGRAMMER:       Matt Wu <mattwu@163.com>
 * HOMEPAGE:         http://ext2.yeah.net
 * UPDATE HISTORY: 
 */

/* INCLUDES *****************************************************************/

#include "ntifs.h"
#include "ext2fs.h"

/* GLOBALS ***************************************************************/

extern PEXT2_GLOBAL Ext2Global;

/* DEFINITIONS *************************************************************/

#ifdef ALLOC_PRAGMA
#pragma alloc_text(PAGE, Ext2LookupFileName)
#pragma alloc_text(PAGE, Ext2SearchFcbList)
#pragma alloc_text(PAGE, Ext2ScanDir)
#pragma alloc_text(PAGE, Ext2OpenFile)
#pragma alloc_text(PAGE, Ext2Create)
#endif


NTSTATUS
Ext2LookupFileName (IN PEXT2_VCB    Vcb,
            IN PUNICODE_STRING      FullFileName,
            IN PEXT2_FCB            ParentFcb,
            OUT PEXT2_MCB *         Ext2Mcb,
            IN OUT PEXT2_INODE      ext2_inode)
{
    NTSTATUS        Status;
    UNICODE_STRING  FileName;
    PEXT2_MCB       ParentMcb;
    PEXT2_MCB       Mcb;

    EXT2_DIR_ENTRY2 ext2_dir;
    int             i = 0;
    BOOLEAN         bRun = TRUE;
    BOOLEAN         bParent = FALSE;
    EXT2_INODE      in;
    ULONG           off = 0;

    Status = STATUS_OBJECT_NAME_NOT_FOUND;

    *Ext2Mcb = NULL;

    if (ParentFcb && FullFileName->Buffer[0] != L'\\')
    {
        ParentMcb = ParentFcb->Ext2Mcb;
        bParent = TRUE;
    }
    else
    {
        ParentMcb = Vcb->Ext2McbTree;
    }

    RtlZeroMemory(&ext2_dir, sizeof(EXT2_DIR_ENTRY2));

    if (FullFileName->Length == 0)
    {
        return Status;
    }

    if (FullFileName->Length == 2 && FullFileName->Buffer[0] == L'\\')
    {
        if (!Ext2LoadInode(Vcb, ParentMcb->Inode, ext2_inode))
        {
            return Status;      
        }

        *Ext2Mcb = Vcb->Ext2McbTree;

        return STATUS_SUCCESS;
    }

    while (bRun && i < FullFileName->Length/2)
    {
        int Length;
        ULONG FileAttr = FILE_ATTRIBUTE_NORMAL;

        if (bParent)
        {
            bParent = FALSE;
        }
        else
        {
            while(i < FullFileName->Length/2 && FullFileName->Buffer[i] == L'\\') i++;
        }

        Length = i;

        while(i < FullFileName->Length/2 && (FullFileName->Buffer[i] != L'\\')) i++;

        if (i - Length >0)
        {
            FileName = *FullFileName;
            FileName.Buffer += Length;
            FileName.Length = (USHORT)((i - Length) * 2);

            Mcb = Ext2SearchMcb(ParentMcb, &FileName);

            if (Mcb)
            {
                ParentMcb = Mcb;

                Status = STATUS_SUCCESS;

                if (!FlagOn(Mcb->FileAttr, FILE_ATTRIBUTE_DIRECTORY))
                {
                    if (i < FullFileName->Length/2)
                    {
                        Status =  STATUS_OBJECT_NAME_NOT_FOUND;
                    }
                    break;
                }
            }
            else
            {
                if (!Ext2LoadInode(Vcb, ParentMcb->Inode, &in))
                {
                    Status = STATUS_OBJECT_NAME_NOT_FOUND;
                    break;
                }

                if (!S_ISDIR(in.i_mode))
                {
                    if (i < FullFileName->Length/2)
                    {
                        Status =  STATUS_OBJECT_NAME_NOT_FOUND;
                    }
                    break;
                }

                Status = Ext2ScanDir (
                    Vcb,
                    ParentMcb,
                    &FileName,
                    &off,
                    &in,
                    &ext2_dir);

                if (!NT_SUCCESS(Status))
                {
                    bRun = FALSE;
                }
                else
                {
                    if (!Ext2LoadInode(Vcb, ext2_dir.inode, &in))
                    {
                        Status = STATUS_OBJECT_NAME_NOT_FOUND;
                        break;
                    }

                    if (S_ISDIR(in.i_mode))
                    {
                        SetFlag(FileAttr, FILE_ATTRIBUTE_DIRECTORY);
                    }

					Mcb = Ext2AllocateMcb(Vcb, &FileName, FileAttr);

					if (!Mcb)
					{
						Status = STATUS_OBJECT_NAME_NOT_FOUND;
						break;
					}

                    Mcb->Inode = ext2_dir.inode;
                    Mcb->DeOffset = off;
                    Ext2AddMcbNode(ParentMcb, Mcb);
                    ParentMcb = Mcb;
                }

            }
        }
        else
        {
            break;
        }
    }

    if (NT_SUCCESS(Status))
    {
        *Ext2Mcb = Mcb;
        if (ext2_inode)
            Ext2LoadInode(Vcb, Mcb->Inode, ext2_inode);
    }
    else
    {
        *Ext2Mcb = NULL;
    }

    return Status;
}


NTSTATUS
Ext2ScanDir (IN PEXT2_VCB       Vcb,
         IN PEXT2_MCB           ParentMcb,
         IN PUNICODE_STRING     FileName,
         IN OUT PULONG          Index,
         IN PEXT2_INODE         ext2_inode,
         IN OUT PEXT2_DIR_ENTRY2 ext2_dir)
{
    NTSTATUS                Status = STATUS_UNSUCCESSFUL;
    USHORT                  InodeFileNameLength;
    UNICODE_STRING          InodeFileName;

    PEXT2_DIR_ENTRY2        pDir = NULL;
    ULONG                   dwBytes = 0;
    BOOLEAN                 bFound = FALSE;
    PEXT2_FCB               Dcb = NULL;
    LONGLONG                Offset = 0;
    ULONG                   inode = ParentMcb->Inode;
    IO_STATUS_BLOCK         IoStatus;   

    InodeFileName.Buffer = NULL;

    __try
    {

        pDir = (PEXT2_DIR_ENTRY2) ExAllocatePool(PagedPool,
                                    sizeof(EXT2_DIR_ENTRY2));
        if (!pDir)
        {
            Status = STATUS_INSUFFICIENT_RESOURCES;
            __leave;
        }

        Dcb = ParentMcb->Ext2Fcb;

        dwBytes = 0;

        while (!bFound && dwBytes < ext2_inode->i_size)
        {
            RtlZeroMemory(pDir, sizeof(EXT2_DIR_ENTRY2));

            // Reading the DCB contents
            if (!(Dcb && CcCopyRead( Dcb->StreamObj, (PLARGE_INTEGER)&Offset,
                            sizeof(EXT2_DIR_ENTRY2), TRUE, (PVOID)pDir,
                            &IoStatus) ) )
            {
                IoStatus.Status = Ext2ReadInode(
                        NULL,
                        Vcb,
                        ext2_inode,
                        dwBytes,
                        (PVOID)pDir,
                        sizeof(EXT2_DIR_ENTRY2),
                        &(IoStatus.Information));
            }

            if (!NT_SUCCESS(IoStatus.Status))
            {
                KdPrint(("Ext2ScanDir: Reading Directory Content error.\n"));
                __leave;
            }

            if (pDir->inode)
            {
                InodeFileNameLength = pDir->name_len & 0xff;
                
                InodeFileName.Length = InodeFileName.MaximumLength =
                    InodeFileNameLength * 2;
                
                InodeFileName.Buffer = ExAllocatePool(
                    PagedPool,
                    InodeFileNameLength * 2 + 2);

                if (!InodeFileName.Buffer)
                {
                    Status = STATUS_INSUFFICIENT_RESOURCES;
                    __leave;
                }

                RtlZeroMemory(InodeFileName.Buffer, InodeFileNameLength * 2 + 2);
                
                Ext2CharToWchar(
                    InodeFileName.Buffer,
                    pDir->name,
                    InodeFileNameLength );

                if (!RtlCompareUnicodeString(
                    FileName,
                    &InodeFileName,
                    TRUE ))
                {
                    bFound = TRUE;
                    *Index = dwBytes;
                    RtlCopyMemory(ext2_dir, pDir, pDir->rec_len > sizeof(EXT2_DIR_ENTRY2)
                        ? sizeof(EXT2_DIR_ENTRY2) : pDir->rec_len);
                    Status = STATUS_SUCCESS;
                    KdPrint(("Ext2ScanDir: Found: Name=%S Inode=%xh\n", InodeFileName.Buffer, pDir->inode));
                }
                
                if (InodeFileName.Buffer != NULL)
                {
                    ExFreePool(InodeFileName.Buffer);
                    InodeFileName.Buffer = NULL;
                }
                
                dwBytes +=pDir->rec_len;
                Offset = (LONGLONG)dwBytes;
            }
            else
            {
                break;
            }
        }

        if (!bFound)
            Status = STATUS_NO_SUCH_FILE;
    }

    __finally
    {
        if (InodeFileName.Buffer != NULL)
        {
            ExFreePool(InodeFileName.Buffer);
        }

        if (pDir)
            ExFreePool(pDir);
    }
    
    return Status;
}

/*

PEXT2_FCB
Ext2SearchFcbList(	IN PEXT2_VCB    Vcb,
				    IN ULONG        inode )
{
	BOOLEAN				bFound = FALSE;
	PLIST_ENTRY         Link;
	PEXT2_FCB           TmpFcb;

	Link = Vcb->FcbList.Flink;
	
	while (!bFound && Link != &Vcb->FcbList)
	{
		TmpFcb = CONTAINING_RECORD(Link, EXT2_FCB, Next);
		
		if (TmpFcb && TmpFcb->Identifier.Type == FCB)
		{
#if DBG
			KdPrint(("Ext2SearchFcbList: [%s,%xh]\n", 
                TmpFcb->AnsiFileName.Buffer, TmpFcb->Inode));
#endif			
			if (TmpFcb->Inode == inode)
			{
				KdPrint(("Ext2SearchMcb: Found FCB for %xh.\n", inode));
				bFound = TRUE;
			}
		}
		Link = Link->Flink;
	}

	if (bFound)
		return TmpFcb;
	else
		return NULL;
	
}
*/

NTSTATUS
Ext2OpenFile(PEXT2_VCB Vcb, PIRP Irp)
{
    NTSTATUS            Status = STATUS_UNSUCCESSFUL;
    PIO_STACK_LOCATION  io_stack;
    PEXT2_FCB           Fcb = NULL;
    PEXT2_FCB           pParentFcb = NULL;
    PEXT2_CCB           Ccb = NULL;
    PEXT2_INODE         ext2_inode;
    BOOLEAN             VcbResourceAcquired = FALSE;
    BOOLEAN             bDir = FALSE;
    BOOLEAN             bCreated = FALSE;
    UNICODE_STRING      FileName;
    PEXT2_MCB           Ext2Mcb = NULL;

    FileName.Buffer = NULL;

    io_stack = IoGetCurrentIrpStackLocation(Irp);

    __try
    {
        ExAcquireResourceExclusiveLite(
            &Vcb->MainResource, TRUE );
        
        VcbResourceAcquired = TRUE;
        
        if (!(ext2_inode = ExAllocatePool   (
              PagedPool, sizeof(EXT2_INODE) ) ))
        {
            __leave;
        }

        FileName.MaximumLength = 0x400;

        if (io_stack->FileObject->RelatedFileObject)
        {
            pParentFcb = (PEXT2_FCB)(io_stack->FileObject->RelatedFileObject->FsContext);
        }

        FileName =  io_stack->FileObject->FileName;

        if ((FileName.Length > sizeof(WCHAR)) &&
            (FileName.Buffer[1] == L'\\') &&
            (FileName.Buffer[0] == L'\\')) {
            
            FileName.Length -= sizeof(WCHAR);
            
            RtlMoveMemory( &FileName.Buffer[0],
                &FileName.Buffer[1],
                FileName.Length );
            
            //
            //  Bad Name if there are still two beginning backslashes.
            //
            
            if ((FileName.Length > sizeof(WCHAR)) &&
                (FileName.Buffer[1] == L'\\') &&
                (FileName.Buffer[0] == L'\\')) {
                                
                Status = STATUS_OBJECT_NAME_INVALID;

                __leave;
            }
        }

        KdPrint(("Ext2OpenFile: %S (name len=%xh)Opt: %xh.\n",
            FileName.Buffer, FileName.Length,
            io_stack->Parameters.Create.Options));

        Status = Ext2LookupFileName(
            Vcb,
            &FileName,
            pParentFcb,
            &Ext2Mcb,
            ext2_inode );

        if (!NT_SUCCESS(Status))
        {
            KdPrint(("Ext2OpenFile: File Not Found.\n"));
            __leave;
        }
        
        Fcb = Ext2Mcb->Ext2Fcb;

        if (!Fcb)
        {
            Fcb = Ext2AllocateFcb (
                    Vcb, &FileName,
                    Ext2Mcb, ext2_inode);
            bCreated = TRUE;
        }
        
        if (Fcb)
        {
            Ccb = Ext2AllocateCcb();

            ExAcquireResourceExclusiveLite(&Fcb->CountResource, TRUE);
            Fcb->OpenHandleCount++;
            Fcb->ReferenceCount++;
            ExReleaseResourceForThreadLite(
                    &Fcb->CountResource,
                    ExGetCurrentResourceThread());

            ExAcquireResourceExclusiveLite(&Vcb->CountResource, TRUE);
            Vcb->OpenFileHandleCount++;
            Vcb->ReferenceCount++;
            ExReleaseResourceForThreadLite(
                    &Vcb->CountResource,
                    ExGetCurrentResourceThread());
            
#if DBG
            KdPrint(("Ext2OpenFile: %s refercount=%xh\n", Fcb->AnsiFileName.Buffer, Fcb->ReferenceCount));
#endif

            if (!FlagOn(Fcb->Ext2Mcb->FileAttr, FILE_ATTRIBUTE_DIRECTORY))
            {
                Fcb->CommonFCBHeader.IsFastIoPossible = FastIoIsPossible;
            }

            io_stack->FileObject->FsContext = (void*)Fcb;
            io_stack->FileObject->FsContext2 = (void*) Ccb;
            io_stack->FileObject->PrivateCacheMap = NULL;
            io_stack->FileObject->SectionObjectPointer = &(Fcb->SectionObject);
            io_stack->FileObject->Vpb = Vcb->Vpb;

            Irp->IoStatus.Information = FILE_OPENED;
            Status = STATUS_SUCCESS;
#if DBG
            KdPrint(("Ext2OpenFile: %s OpenCount: %u ReferCount: %u\n",
                Fcb->AnsiFileName.Buffer, Fcb->OpenHandleCount, Fcb->ReferenceCount));
#endif
        }
        else
        {
            Irp->IoStatus.Information = 0;
            Status = STATUS_OBJECT_NAME_NOT_FOUND;
            KdPrint(("Ext2OpenFile: Create denied (CCB allocate error.)\n"));
        }
    
    }

    __finally
    {

        if (VcbResourceAcquired)
        {
            ExReleaseResourceForThreadLite(
                &Vcb->MainResource,
                ExGetCurrentResourceThread() );
        }

        if (!bCreated)
        {
            if (ext2_inode)
                ExFreePool(ext2_inode);
        }
        else
        {
            if (!Fcb && ext2_inode)
                ExFreePool(ext2_inode);
        }
    }
    
    return Status;
}


NTSTATUS
Ext2Create (IN PEXT2_IRP_CONTEXT IrpContext)
{
    PDEVICE_OBJECT      DeviceObject;
    PIRP                Irp;
    PIO_STACK_LOCATION  io_stack;
    PEXT2_VCB           Vcb = 0;
    NTSTATUS            Status = STATUS_OBJECT_NAME_NOT_FOUND;

    DeviceObject = IrpContext->DeviceObject;

    Vcb = (PEXT2_VCB) DeviceObject->DeviceExtension;
    
    Irp = IrpContext->Irp;
    
    io_stack = IoGetCurrentIrpStackLocation(Irp);
    
    if (DeviceObject == Ext2Global->DeviceObject)
    {
        KdPrint(("Ext2Create: Create on main device object.\n"));
        
        Irp->IoStatus.Information = FILE_OPENED;
        Status = STATUS_SUCCESS;
        IrpContext->Irp->IoStatus.Status = Status;
        
        Ext2CompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
        
        Ext2FreeIrpContext(IrpContext);
        
        return Status;
        
    }
    else if (io_stack->FileObject->FileName.Length == 0)
    {
        io_stack->FileObject->FsContext = Vcb;
        
        ExAcquireResourceExclusiveLite(
            &Vcb->CountResource, TRUE );
        
        Vcb->ReferenceCount++;
        
        ExReleaseResourceForThreadLite(
            &Vcb->CountResource,
            ExGetCurrentResourceThread() );
        
        Irp->IoStatus.Information = FILE_OPENED;
        Status = STATUS_SUCCESS;
        IrpContext->Irp->IoStatus.Status = Status;
        
        Ext2CompleteRequest(IrpContext->Irp, IO_NO_INCREMENT);
        
        Ext2FreeIrpContext(IrpContext);
        
        return Status;
    }
    
    __try
    {
        Status = Ext2OpenFile(Vcb, Irp);
    }

    __finally
    {

        if (!IrpContext->ExceptionInProgress)
        {
            IrpContext->Irp->IoStatus.Status = Status;
            
            Ext2CompleteRequest(
                IrpContext->Irp,
                (CCHAR)
                (NT_SUCCESS(Status) ? IO_DISK_INCREMENT : IO_NO_INCREMENT));
            
            Ext2FreeIrpContext(IrpContext);

            // ???? - Strange action here - May. 23
/*          
            if (Vcb && (!Vcb->ReferenceCount) &&
                FlagOn(Vcb->Flags, VCB_DISMOUNT_PENDING) )
            {
                Ext2FreeVcb(Vcb);
            }
*/
        }
    }
    
    return Status;
}
